@@ -25,7 +25,9 @@ from html5lib import treewalkers
 from html5lib import treebuilders
 from html5lib.filters import sanitizer
 
-from repoze.lru import lru_cache
+from repoze.lru import CacheMaker
+
+from html2text import html2text
 
 from nti.contentfragments.interfaces import IAllowedAttributeProvider
 
@@ -39,6 +41,34 @@ except NameError:
 Element = getattr(etree, 'Element')
 etree_tostring = getattr(etree, 'tostring')
 
+# This regex is basen on one from Python 3's html.parser module;
+# it is used in the parser's ``check_for_whole_start_tag`` method.
+# That regex in turn comes from the HTML5 specification.
+locatestarttagend_tolerant = re.compile(r"""
+  <[a-zA-Z][^\t\n\r\f />\x00]*       # tag name
+  (?:[\s/]*                          # optional whitespace before attribute name
+    (?:(?<=['"\s/])[^\s/>][^\s/=>]*  # attribute name
+      (?:\s*=+\s*                    # value indicator
+        (?:'[^']*'                   # LITA-enclosed value
+          |"[^"]*"                   # LIT-enclosed value
+          |(?!['"])[^>\s]*           # bare value
+         )
+        \s*                          # possibly followed by a space
+       )?(?:\s|/(?!>))*
+     )*
+   )?
+  \s*                                # trailing whitespace
+""", re.VERBOSE)
+# Likewise, these come from the same module. The detect ``&nbsp;``
+# and friends.
+entityref = re.compile(u'&([a-zA-Z][-.a-zA-Z0-9]*)[^a-zA-Z0-9]')
+charref = re.compile(u'&#(?:[0-9]+|[xX][0-9a-fA-F]+)[^0-9a-fA-F]')
+# TODO: It would be nice (perhaps a performance win?) to be able to combine
+# the three expressions into a single expression. The naive way, simply
+# concatenating them with ``(...)|(...)|(...)``, doesn't work (probably
+# because of the VERBOSE flag)
+
+
 # serializer.xhtmlserializer.XHTMLSerializer is removed in html5lib 1.0.
 # It was simply defined as:
 #
@@ -108,6 +138,22 @@ sanitizer.re = FakeRe()
 
 from html5lib.constants import namespaces
 
+
+@interface.implementer(IHyperlinkFormatter)
+class _NoopHyperlinkFormatter(object):
+    """
+    A fallback hyperlink formatter that does not attempt to identify
+    links in the provided text.
+    """
+
+    def find_links(self, text):
+        return (text, )
+
+    def format(self, html_fragment):
+        # TODO should this just return html_fragment?
+        raise NotImplementedError # pragma: no cover
+
+
 # But we define our own sanitizer mixin subclass and filter to be able to
 # customize the allowed tags and protocols
 class _SanitizerFilter(sanitizer.Filter):
@@ -117,7 +163,8 @@ class _SanitizerFilter(sanitizer.Filter):
 
     def __init__(self, *args, **kwargs):
         super(_SanitizerFilter, self).__init__(*args, **kwargs)
-        self.link_finder = component.queryUtility(IHyperlinkFormatter)
+        self.link_finder = component.queryUtility(IHyperlinkFormatter,
+                                                  default=_NoopHyperlinkFormatter())
 
         acceptable_elements = frozenset([
             'a', 'audio',
@@ -195,7 +242,7 @@ class _SanitizerFilter(sanitizer.Filter):
 
     def _find_links_in_text(self, token):
         text = token['data']
-        text_and_links = self.link_finder.find_links(text) if self.link_finder else ()
+        text_and_links = self.link_finder.find_links(text)
         if len(text_and_links) != 1 or text_and_links[0] != text:
 
             def _unicode(x):
@@ -301,7 +348,7 @@ def _to_sanitized_doc(user_input):
     # <html>...</html> => <html><body>&lthtml&gt...&lt/html&gt</html>
     __traceback_info__ = user_input
     p = HTMLParser(tree=treebuilders.getTreeBuilder("lxml"),
-                    namespaceHTMLElements=False)
+                   namespaceHTMLElements=False)
     doc = p.parse(user_input)
     string = _html5lib_tostring(doc, sanitize=True)
 
@@ -315,10 +362,6 @@ def _to_sanitized_doc(user_input):
     doc = p.parse(string)
     return doc
 
-def _doc_to_plain_text(doc):
-    string = etree_tostring(doc, method='text', encoding=unicode)
-    return PlainTextContentFragment(string)
-
 def sanitize_user_html(user_input, method='html'):
     """
     Given a user input string of plain text, HTML or HTML fragment, sanitize
@@ -329,12 +372,22 @@ def sanitize_user_html(user_input, method='html'):
         :func:`lxml.etree.tostring`. The default value, ``html``, causes this
         method to produce either HTML or plain text, whatever is most appropriate.
         Passing the value ``text`` causes this method to produce only plain text captured
-        by traversing the elements with lxml.
+        by traversing the elements with lxml. Note: this is legacy functionality,
+        and callers should generally convert via calling the interfaces.
 
     :return: Something that implements :class:`frg_interfaces.IUnicodeContentFragment`,
         typically either :class:`frg_interfaces.IPlainTextContentFragment` or
         :class:`frg_interfaces.ISanitizedHTMLContentFragment`.
     """
+    # Registered as the adapter from (bytes/unicode) -> IUnicodeContentFragment
+    # And as the adapter from IHTMLContentFragment -> ISanitizedHTMLContentFragment
+    # (even though that may be plain text).
+
+    if not may_contain_html_like_markup(user_input):
+        return _sanitize_user_html_to_text(user_input, _guaranteed_no_markup=True)
+
+    if method == 'text':
+        return _sanitize_user_html_to_text(user_input)
 
     doc = _to_sanitized_doc(user_input)
 
@@ -375,9 +428,6 @@ def sanitize_user_html(user_input, method='html'):
             new_style = style + (' ' if style else '') + 'max-width: 100%;'
             node.attrib['style'] = new_style
 
-    if method == 'text':
-        return _doc_to_plain_text(doc)
-
     string = _html5lib_tostring(doc, sanitize=False)
     # If we can go back to plain text, do so.
     normalized = string[len('<html><head></head><body>'): 0 - len('</body></html>')]
@@ -386,10 +436,13 @@ def sanitize_user_html(user_input, method='html'):
         normalized = normalized[0:-6]
 
     # If it has no more tags, we can be plain text.
-    # TODO: This probably breaks on entities like &lt;
-    if '<' not in normalized:
-        # 03.2016 - JZ - We no longer want to lstrip.
-        string = PlainTextContentFragment(normalized.rstrip())
+    if not may_contain_html_like_markup(normalized):
+        # Going via the to-markdown converter yields prettier results than simply
+        # returning a PlainTextContentFragment. This does mean we parse it twice,
+        # but html2text is fast.
+        # TODO: does this suffer from issue 44, notably, being too aggressive about
+        # some conversions?
+        string = _sanitize_user_html_to_text(user_input)
     else:
         string = SanitizedHTMLContentFragment(u"<html><body>" + normalized + u"</body></html>")
     return string
@@ -401,39 +454,98 @@ def sanitize_user_html(user_input, method='html'):
 # in a non-persistent field of the object? (But the objects aren't Persistent, so
 # _v fields might not work?)
 
+_cache_maker = CacheMaker(10000)
+
+def _ensure_unicode_non_lossy(user_input):
+    # Decode to unicode using an encoding that doesn't lose any bytes;
+    # let the HTML parser deal with anything at a higher level.
+    if isinstance(user_input, bytes):
+        user_input = user_input.decode('latin-1')
+    return user_input
+
+@_cache_maker.lrucache()
+def may_contain_html_like_markup(user_input):
+    # Detect if there are valid start tags or entity/character references that should
+    # be stripped/replaced.
+
+    # Python 2 can match bytes patterns against unicode or bytes input, and vice versa,
+    # unicode patterns against bytes or unicode input. However, Python 3 only works
+    # when both pattern and input are the same type. The patterns will be unicode on Python 3,
+    # so make sure our input is as well.
+    user_input = _ensure_unicode_non_lossy(user_input)
+
+    # search() -> match anywhere in the string; match() -> only the beginning
+
+    return (
+        locatestarttagend_tolerant.search(user_input)
+        or entityref.search(user_input)
+        or charref.search(user_input)
+    )
+
+
 @interface.implementer(IPlainTextContentFragment)
 @component.adapter(basestring)
-@lru_cache(10000)
-def _sanitize_user_html_to_text(user_input):
+@_cache_maker.lrucache()
+def _sanitize_user_html_to_text(user_input, _guaranteed_no_markup=False):
     """
+    _sanitize_user_html_to_text(user_input) -> str
+
     Registered as an adapter with the name 'text' for convenience.
 
     See :func:`sanitize_user_html`.
+
+    .. caution::
+       While this adapter accepts an arbitrary base string, it does not actually
+       guarantee the output is totally plain text suitable for any purpose.
+       In particular, it only attempts to extract human-readable text from things
+       that look like HTML markup while preserving most information such as links.
+       It does not attempt to extract human-readable text from things like
+       LaTeX or ReST input; input in forms like that may be returned unaltered or
+       altered beyond recognition.
+
+       This adapter also does not attempt to escape any characters that may have
+       special meaning to HTML, such as ``<`` if the input does not otherwise appear to be
+       HTML.
+
     """
     # We are sometimes used as a named adapter, or even sadly called
     # directly, which means we can get called even with the right kind
     # of input already. It messes the content up if we try to reparse
     if IPlainTextContentFragment.providedBy(user_input):
         return user_input
-    __traceback_info__ = user_input, type(user_input)
 
-    return sanitize_user_html(user_input, method='text')
+    # Decode to unicode using a sequence that doesn't lose any bytes;
+    # let the HTML parser deal with anything at a higher level.
+    user_input = _ensure_unicode_non_lossy(user_input)
+
+    if _guaranteed_no_markup or not may_contain_html_like_markup(user_input):
+        output = user_input
+    else:
+        # Using a wider bodywidth helps prevent unneeded line breaks,
+        # especially in tests.
+        output = html2text(user_input, bodywidth=100)
+
+    # The old lxml-based implementation stripped trailing whitespace on each line; for compatibility
+    # do the same
+    output = '\n'.join(l.rstrip() for l in output.splitlines())
+    output = output.rstrip()
+    return PlainTextContentFragment(output)
 
 @interface.implementer(IPlainTextContentFragment)
 @component.adapter(IHTMLContentFragment)
-@lru_cache(10000)
+@_cache_maker.lrucache()
 def _html_to_sanitized_text(html):
-    if IPlainTextContentFragment.providedBy(html):
-        return html
-    __traceback_info__ = html, type(html)
-
-    return _doc_to_plain_text(_to_sanitized_doc(html))
+    return _sanitize_user_html_to_text(html)
 
 @interface.implementer(IPlainTextContentFragment)
 @component.adapter(ISanitizedHTMLContentFragment)
-@lru_cache(10000)
+@_cache_maker.lrucache()
 def _sanitized_html_to_sanitized_text(sanitized_html):
-    p = HTMLParser(tree=treebuilders.getTreeBuilder("lxml"),
-                   namespaceHTMLElements=False)
-    doc = p.parse(sanitized_html)
-    return _doc_to_plain_text(doc)
+    return _sanitize_user_html_to_text(sanitized_html)
+
+try:
+    from zope.testing.cleanup import addCleanUp
+except ImportError: # pragma: no cover
+    pass
+else:
+    addCleanUp(_cache_maker.clear)
